from __future__ import annotations

from typing import *  # type: ignore

from . import component_base

__all__ = ["Container"]


class Container(component_base.Component):
    """
    A component holding a single child.

    `Container` is a simple container which holds a single child component. It
    is useful for when you receive a component as child and wish to add
    additional layout attributes such as a margin.

    Attributes:
        child: The component to place inside the container.
    """

    child: component_base.Component

    def build(self) -> component_base.Component:
        return self.child



from __future__ import annotations

from typing import *  # type: ignore

import rio

from . import class_container, component_base

__all__ = [
    "Overlay",
]


class Overlay(component_base.Component):
    """
    Displays its child above all other components.

    The overlay component takes a single child widget, and displays it above all
    other components on the page. The child will not scroll with the rest of the
    page and is exempt from layouting.

    Components inside of overlays are allocated the entire screen, and are
    themselves responsible for positioning themselves as required. You can
    easily achieve this using the child's `align_x` and `align_y` properties.

    Attributes:
        child: The component to display in the overlay. It will take up the
        entire size of the screen, so make sure to use properties such as
        `align_x` and `align_y` to position it as needed.
    """

    def __init__(
        self,
        child: rio.Component,
        *,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["natural", "grow"], float] = "natural",
        height: Union[Literal["natural", "grow"], float] = "natural",
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        # Passing the layout values through to `Component` would ignore them,
        # because the HTML `Overlay` element will force itself to span the entire
        # screen.
        #
        # Instead, simply keep track of the values, then pass them on to a
        # `Container` child.
        #
        # Furthermore, another, outer `Container` is needed to ensure that the
        # layouting system doesn't apply any problematic CSS attributes to the
        # `Overlay` div.

        super().__init__(
            key=key,
        )
        self._passthrough_margin = margin
        self._passthrough_margin_x = margin_x
        self._passthrough_margin_y = margin_y
        self._passthrough_margin_left = margin_left
        self._passthrough_margin_top = margin_top
        self._passthrough_margin_right = margin_right
        self._passthrough_margin_bottom = margin_bottom
        self._passthrough_width = width
        self._passthrough_height = height
        self._passthrough_align_x = align_x
        self._passthrough_align_y = align_y

        self.child = child

    def build(self) -> rio.Component:
        # Outer: This container absorbs any CSS attributes assigned by the
        # layouting system, so that they don't interfere with the `Overlay`
        # element.
        return rio.Container(
            # Overlay: This is the actual overlay element. It spans the entire
            # screen and doesn't scroll.
            class_container.ClassContainer(
                # Inner: This container is used to apply any layouting
                # attributes passed to the `Overlay` component.
                rio.Container(
                    margin=self._passthrough_margin,
                    margin_x=self._passthrough_margin_x,
                    margin_y=self._passthrough_margin_y,
                    margin_left=self._passthrough_margin_left,
                    margin_top=self._passthrough_margin_top,
                    margin_right=self._passthrough_margin_right,
                    margin_bottom=self._passthrough_margin_bottom,
                    width=self._passthrough_width,  # type: ignore
                    height=self._passthrough_height,  # type: ignore
                    align_x=self._passthrough_align_x,
                    align_y=self._passthrough_align_y,
                    child=rio.Container(
                        child=self.child,
                    ),
                ),
                classes=[
                    "rio-overlay",
                ],
            ),
        )



import logging
import zipfile
from pathlib import Path
from typing import *  # type: ignore

from . import common
from .errors import AssetError

_icon_registry: Optional["IconRegistry"] = None


class IconRegistry:
    """
    Helper class, which keeps track of all icons known to rio.

    Icons are stored in a zip file, and extracted on first use. Any icons which
    have been accessed are kept in memory for rapid access.
    """

    def __init__(self, cache_dir: Path):
        self.cache_dir = cache_dir

        # Maps icon names (set/icon:variant) to the icon's SVG string. The icon
        # names are canonical form.
        self.cached_icons: Dict[str, str] = {}

        # Maps icon set names to the path of the zip file containing the icons
        self.icon_set_archives: Dict[str, Path] = {}

    @classmethod
    def get_singleton(cls) -> "IconRegistry":
        """
        Return the singleton instance of this class.
        """
        global _icon_registry

        # Create the instance if there is none yet
        if _icon_registry is None:
            _icon_registry = IconRegistry(common.RIO_CACHE_DIR / "extracted-icon-sets")

            # Register built-in icon sets
            _icon_registry.icon_set_archives["material"] = (
                common.RIO_ASSETS_DIR / "compressed-icon-sets" / "material.zip"
            )

        # Use it
        return _icon_registry

    @staticmethod
    def parse_icon_name(icon_name: str) -> Tuple[str, str, Optional[str]]:
        """
        Given a name for an icon, return the three parts of the name: set, icon,
        variant. If the name is syntactically invalid (e.g. too many slashes),
        raise an `AssetError`.
        """
        # Determine the icon set
        sections = icon_name.split("/")

        if len(sections) == 1:
            icon_set = "material"
            icon_name = sections[0]
        elif len(sections) == 2:
            icon_set, icon_name = sections
        else:
            raise AssetError(
                f"Invalid icon name `{icon_name}`. Icons names must be of the form `set/icon:variant`"
            )

        # Determine the icon name and variant
        sections = icon_name.split(":")

        if len(sections) == 1:
            return icon_set, sections[0], None

        if len(sections) == 2:
            return icon_set, sections[0], sections[1]

        raise AssetError(
            f"Invalid icon name `{icon_name}`. Icons names must be of the form `set/icon:variant`"
        )

    @staticmethod
    def normalize_icon_name(icon_name: str) -> str:
        """
        Given a name for an icon, return the canonical form of the name. If the
        name is syntactically invalid (e.g. too many slashes), raise an
        `AssetError`.
        """

        set, name, section = IconRegistry.parse_icon_name(icon_name)

        if section is None:
            return f"{set}/{name}"

        return f"{set}/{name}:{section}"

    def _get_icon_svg_path(
        self,
        icon_name: str,
    ) -> Path:
        """
        Given an icon name, return the path to the SVG file for that icon. This
        will extract the icon  if necessary.
        """
        # Prepare some paths
        icon_set, icon_name, variant = IconRegistry.parse_icon_name(icon_name)

        icon_set_dir = self.cache_dir / icon_set

        if variant is None:
            svg_path = icon_set_dir / f"{icon_name}.svg"
        else:
            svg_path = icon_set_dir / variant / f"{icon_name}.svg"

        # Check if the file has already been extracted
        if svg_path.exists():
            return svg_path

        # If the set's directory already exists there is nothing to do - this is
        # simply an invalid icon name
        if icon_set_dir.exists():
            return svg_path

        # If there is no icon set matching the name there is also nothing to do,
        # as this is an invalid / unregistered icon set
        try:
            archive_path = self.icon_set_archives[icon_set]
        except KeyError:
            return svg_path

        # The cache directory doesn't exist. Extract the icon set's zip file to
        # create it
        target_dir = self.cache_dir / icon_set
        target_dir.mkdir(parents=True, exist_ok=True)

        logging.debug(
            f"Extracting icon set `{icon_set}` from `{archive_path}` to `{target_dir}`"
        )

        with zipfile.ZipFile(archive_path, "r") as zip_file:
            zip_file.extractall(target_dir)

        return svg_path

    def get_icon_svg(self, icon_name: str) -> str:
        """
        Given an icon name, return the SVG string for that icon. This will
        extract the icon if necessary. If the icon name is invalid or there is
        no matching icon, raise an `AssetError`.
        """

        # Normalize the icon name
        icon_name = IconRegistry.normalize_icon_name(icon_name)

        # Already cached?
        try:
            return self.cached_icons[icon_name]
        except KeyError:
            pass

        # Get the path to the icon's SVG file
        svg_path = self._get_icon_svg_path(icon_name)

        print(icon_name, svg_path)

        # Read the SVG file
        try:
            svg_string = svg_path.read_text()
        except FileNotFoundError:
            # Figure out which part of the name is the problem to show a
            # descriptive error message
            icon_set, icon_name, variant = IconRegistry.parse_icon_name(icon_name)

            icon_set_path = self.cache_dir / icon_set

            if not icon_set_path.exists():
                raise AssetError(
                    f"Unknown icon set `{icon_set}`. Known icon sets are: `{'`, `'.join(self.icon_set_archives.keys())}`"
                )

            raise AssetError(
                f"There is no icon named `{icon_name}` in the `{icon_set}` icon set"
            )

        # Cache the icon
        self.cached_icons[icon_name] = svg_string

        # Done
        return svg_string



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from . import component_base

__all__ = [
    "MarkdownView",
]


class MarkdownView(component_base.FundamentalComponent):
    """
    Displays Markdown-formatted text.

    `MarkdownView` displays text formatted using Markdown. Markdown is a
    lightweight markup language that allows you to write text with simple
    formatting, such as **bold**, *italics*, and links.

    Markdown is a great way to write text that is both human-readable, yet
    beautifully formatted.

    Attributes:
        text: The Markdown-formatted text to display.

        default_language: The default language to use for code blocks. If
            `None`, Rio will try to guess the language automatically. If a
            default is given, it will be used for all code blocks that don't
            specify a language explicitly.

            Inline code will always use the default language, since they are too
            short to reliably guess the language - so make sure to set a default
            language if you want your inline code to be syntax-highlighted.
    """

    text: str
    _: KW_ONLY
    default_language: Optional[str] = None


MarkdownView._unique_id = "MarkdownView-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "Popup",
    "PopupOpenOrCloseEvent",
]


@dataclass
class PopupOpenOrCloseEvent:
    is_open: bool


class Popup(component_base.FundamentalComponent):
    """
    A container which floats above other components.

    Popups are containers which float above the page when open. This allows you
    to keep your app clean by default, but present additional information or
    controls when needed.

    They take two children: The `anchor` is always visible and positions the
    popup. The `content` is located inside the popup and is only visible when
    the popup is open.

    The location popups appear at can be customized using the `direction`,
    `alignment` and `gap` attributes. Popups wil do their best to honor those
    settings, but deviate if necessary to ensure they don't go off-screen.

    Attributes:
        anchor: A component which is always visible and positions the popup.

        content: A component which is only visible when the popup is open.

        direction: The direction into which the popup opens.

        alignment: The alignment of the popup within the anchor. If the popup
            opens to the left or right, this is the vertical alignment, with `0`
            being the top and `1` being the bottom. If the popup opens to the
            top or bottom, this is the horizontal alignment, with `0` being the
            left and `1` being the right. Has no effect if the popup opens
            centered.

        gap: How much space to leave between the popup and the anchor. Has no
            effect popup opens centered.

        is_open: Whether the popup is currently open.
    """

    anchor: rio.Component
    content: rio.Component
    _: KW_ONLY
    direction: Literal["left", "top", "right", "bottom", "center"] = "center"
    alignment: float = 0.5
    gap: float = 0.0
    is_open: bool = False
    on_open_or_close: rio.EventHandler[PopupOpenOrCloseEvent] = None

    async def _on_state_update(self, delta_state: JsonDoc) -> None:
        # Trigger on_open_or_close event
        try:
            new_value = delta_state["is_open"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, bool), new_value
            await self.call_event_handler(
                self.on_open_or_close,
                PopupOpenOrCloseEvent(new_value),
            )

        # Chain up
        await super()._on_state_update(delta_state)


Popup._unique_id = "Popup-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from . import component_base

__all__ = [
    "ProgressCircle",
]


class ProgressCircle(component_base.FundamentalComponent):
    """
    A progress indicator in the shape of a circle.

    `ProgressCircle` conveys to the user that activity is ongoing. It can either
    display the exact progress as a fraction from 0 to 1, or it can display an
    indeterminate progress animation, which is useful when the exact progress
    isn't known.

    The linear counterpart to this component is the `ProgressBar`.

    Attributes:
        progress: The progress to display, as a fraction from 0 to 1. If `None`,
            the progress indicator will be indeterminate.

        color: The color scheme of the progress indicator. Keeping the default
                is recommended, but it may make sense to change the color in
                case the default is hard to perceive on your background.
    """

    progress: Optional[float]
    color: rio.ColorSet

    def __init__(
        self,
        *,
        progress: Optional[float] = None,
        color: rio.ColorSet = "keep",
        size: Union[Literal["grow"], float] = 3.5,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        """
        Args:
            progress: The progress to display, as a fraction from 0 to 1. If `None`,
                the progress indicator will be indeterminate.

            color: The color scheme of the progress indicator. Keeping the default
                is recommended, but it may make sense to change the color in case
                the default is hard to perceive on your background.

            size: The size of the progress indicator. This is equivalent to setting
                a component's `width` and `height` to the same value.

                Note that unlike most components in Rio, `ProgressCircle` does not
                have a `natural` size, since the circle can easily be scaled to fit
                any size. Therefore it defaults to a reasonable size which should
                fit most use cases.
        """
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=size,
            height=size,
            align_x=align_x,
            align_y=align_y,
        )

        self.progress = progress
        self.color = color


ProgressCircle._unique_id = "ProgressCircle-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .. import color
from . import component_base

__all__ = [
    "ColorPicker",
    "ColorChangeEvent",
]


@dataclass
class ColorChangeEvent:
    color: color.Color


class ColorPicker(component_base.FundamentalComponent):
    """
    Allows the user to pick a RGB(A) color.

    `ColorPicker` is a component that allows the user to pick a color. It
    displays a combination of colorful areas and sliders that the user can
    interact with to pick a color, and optionally an opacity slider to pick
    opacity.

    Attributes:
        color: The color that the user has picked.

        pick_opacity: Whether to allow the user to pick opacity. If `False`,
            the opacity slider will be hidden and the color value will be forced
            to be fully opaque.

        on_change: This event is triggered whenever the user changes the color.
    """

    color: color.Color
    _: KW_ONLY
    pick_opacity: bool = False
    on_change: rio.EventHandler[ColorChangeEvent] = None

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        # Update the color
        self.color = color.Color.from_rgb(*msg["color"])

        # Trigger the change event
        await self.call_event_handler(
            self.on_change,
            ColorChangeEvent(self.color),
        )

        # Refresh the session
        await self.session._refresh()

    def _custom_serialize(self) -> JsonDoc:
        return {
            "color": self.color.rgba,
        }


ColorPicker._unique_id = "ColorPicker-builtin"



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Iterable, Literal, Tuple, Union

from typing_extensions import TypeAlias
from uniserde import Jsonable

import rio

from . import assets, self_serializing, session
from .color import Color
from .common import ImageLike

__all__ = [
    "Fill",
    "FillLike",
    "ImageFill",
    "LinearGradientFill",
    "SolidFill",
]


class Fill(self_serializing.SelfSerializing, ABC):
    """
    Base class for how shapes are filled.

    This is a base class for all fills. Fills determine how the inside of a
    shape is colored.

    This class is abstract and cannot be instantiated directly. Instead, use
    one of its subclasses.
    """

    @staticmethod
    def _try_from(value: FillLike) -> "Fill":
        """
        Creates a fill instance from a `FillLike` value.

        Raises:
            TypeError: If the value is not a valid fill.
        """
        if isinstance(value, Fill):
            return value

        if isinstance(value, Color):
            return SolidFill(value)

        raise TypeError(f"Expected Fill or Color, got {type(value)}")


@dataclass(frozen=True, eq=True)
class SolidFill(Fill):
    """
    Fills a shape with a single color.

    `SolidFill` is the simplest of all fills. It fills the entire shape with a
    single, solid color.

    Attributes:
        color: The color to fill the shape with.
    """

    color: Color

    def _serialize(self, sess: session.Session) -> Jsonable:
        return {
            "type": "solid",
            "color": self.color.rgba,
        }


@dataclass(frozen=True, eq=True)
class LinearGradientFill(Fill):
    """
    Fills a shape with a linear gradient.

    `LinearGradientFill` fills the shape with a linear gradient. The gradient
    can have any number of stops, each with a color and a position. The gradient
    will smoothly transition between the colors at the given positions. The
    positions are given as are given as fractions, where 0 is the start of the
    gradient and 1 is the end.

    Attributes:
        stops: The different colors that comprise the gradient, along with where
            they are positioned.

            The stops are given as tuples. Each tuple contains a color and a
            position. The position is a fraction, where 0 is the start of the
            gradient and 1 is the end.

            The order of the stops has no effect.

            There must be at least one stop.

        angle_degrees: The angle of the gradient, in degrees. 0 degrees points
            straight to the right, and the angle increases counterclockwise.
            (This lines up with how angles are typically used mathematically.)
    """

    stops: Tuple[Tuple[Color, float], ...]
    angle_degrees: float = 0.0

    def __init__(
        self,
        *stops: Tuple[Color, float],
        angle_degrees: float = 0.0,
    ):
        # Make sure there's at least one stop
        if not stops:
            raise ValueError("Gradients must have at least 1 stop")

        # Sort and store the stops
        vars(self).update(
            stops=tuple(sorted(stops, key=lambda x: x[1])),
            angle_degrees=angle_degrees,
        )

    def _as_css_background(self, sess: rio.Session) -> str:
        # Special case: Just one color
        if len(self.stops) == 1:
            return f"#{self.stops[0][0].hex}"

        # Proper gradient
        stop_strings = []

        for stop in self.stops:
            color = stop[0]
            position = stop[1]
            stop_strings.append(f"#{color.hex} {position * 100}%")

        return (
            f"linear-gradient({90 - self.angle_degrees}deg, {', '.join(stop_strings)})"
        )

    def _serialize(self, sess: session.Session) -> Jsonable:
        return {
            "type": "linearGradient",
            "stops": [(color.rgba, position) for color, position in self.stops],
            "angleDegrees": self.angle_degrees,
        }


class ImageFill(Fill):
    """
    Fills a shape with an image.

    `ImageFill` fills the shape's background with an image.

    The image can be scaled to fit the shape in one of three ways:

    - `fit`: The image is scaled to fit entirely inside the shape, while
      maintaining its aspect ratio. This is the default.
    - `stretch`: The image is stretched to fill the shape, distorting it if
      necessary.
    - `zoom`: The image is scaled to fill the shape entirely, while maintaining
      its aspect ratio. This may cause the image to overflow the shape.
    """

    def __init__(
        self,
        image: ImageLike,
        *,
        fill_mode: Literal["fit", "stretch", "zoom"] = "fit",
    ):
        """
        Args:
            image: The image to fill the shape with. fill_mode: How the image should
                be scaled to fit the shape.

            fill_mode: How the image should be scaled to fit the shape. If `fit`,
                the image is scaled to fit entirely inside the shape. If `stretch`,
                the image is stretched to fill the shape exactly, possibly
                distorting it in the process. If `zoom`, the image is scaled to fill
                the shape entirely, possibly overflowing.
        """
        self._image_asset = assets.Asset.from_image(image)
        self._fill_mode = fill_mode

    def _serialize(self, sess: session.Session) -> Jsonable:
        return {
            "type": "image",
            "fillMode": self._fill_mode,
            "imageUrl": self._image_asset._serialize(sess),
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ImageFill):
            return NotImplemented

        return (
            self._image_asset == other._image_asset
            and self._fill_mode == other._fill_mode
        )

    def __hash__(self) -> int:
        return hash((self._image_asset, self._fill_mode))

    def _as_css_background(self, sess: rio.Session) -> str:
        # Fetch the escaped URL. That way it cannot interfere with the CSS syntax
        self._image_asset._serialize(sess)
        image_url = str(self._image_asset.url)
        css_url = f"url('{image_url}')"

        if self._fill_mode == "fit":
            return f"{css_url} center/contain no-repeat"
        elif self._fill_mode == "stretch":
            return f"{css_url} top left / 100% 100%"
        elif self._fill_mode == "tile":
            return f"{css_url} left top repeat"
        elif self._fill_mode == "zoom":
            return f"{css_url} center/cover no-repeat"
        else:
            # Invalid fill mode
            raise Exception(f"Invalid fill mode for image fill: {self._fill_mode}")


FillLike: TypeAlias = Union[Fill, Color]



from __future__ import annotations

from dataclasses import KW_ONLY
from pathlib import Path
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from .. import color, fills, icon_registry
from . import component_base

__all__ = [
    "Icon",
]


class Icon(component_base.FundamentalComponent):
    """
    Displays one of many pre-bundled icons.

    Icons are a great way to add polish to your app. A good icon can help your
    users understand your app and immediately recognize what a component does.

    Rio includes hundreds of free icons, allowing you to easily add them to your
    app without having to find or create your own. The `Icon` component displays
    one of these icons.

    Note that unlike most components in Rio, `Icon` does not have a `natural`
    size, they can be easily be scaled to fit any size. Therefore it defaults to
    a width and height of 1.3, which is a great size when mixing icons with
    text.

    Icon names are in the format `set_name/icon_name:variant`. Rio already ships
    with the `material` icon set, which contains icons in the style of Google's
    Material Design. You can browse all available icons on [the Rio website](
    https://todo.com/icons).  # TODO: Add actual link

    The set name and variant can be omitted. If no set name is specified, it
    defaults to `material`. If not variant is specified, the default version of
    the icon, i.e. no variant, is used.

    Attributes:
        icon: The name of the icon to display, in the format
            `icon-set/name:variant`. You can browse all available icons on [the
            Rio website](https://todo.com/icons).  # TODO: Add actual link

        fill: The color scheme of the icon. The text color is used if no fill is
            specified.
    """

    icon: str
    _: KW_ONLY
    fill: Union[rio.FillLike, color.ColorSet, Literal["dim"]]

    @staticmethod
    def _get_registry() -> icon_registry.IconRegistry:
        return icon_registry.IconRegistry.get_singleton()

    @staticmethod
    def register_icon_set(
        set_name: str,
        icon_set_zip_path: Path,
    ) -> None:
        """
        Add an icon set to the global registry. This allows the icons to be
        accessed as `icon_name`, `set_name/icon_name` or
        `set_name/icon_name:variant`.

        There must not already be a set with the given name.

        The icon set is a zip archive containing SVG files. The SVG files must
        have a `viewBox` attribute, but no height or width. They will be colored
        by the `fill` property of the `Icon` component.

        Files located in the root of the archive can be accessed as
        `set_name/icon_name`. Files located in a subdirectory can be accessed as
        `set_name/icon_name:variant`.

        Args:
            set_name: The name of the new icon set. This will be used to access
                the icons.

            icon_set_zip_path: The path to the zip archive containing the icon
                set.
        """
        registry = Icon._get_registry()

        if set_name in registry.icon_set_archives:
            raise ValueError(f"There is already an icon set named `{set_name}`")

        registry.icon_set_archives[set_name] = icon_set_zip_path

    def __init__(
        self,
        icon: str,
        *,
        fill: Union[rio.FillLike, color.ColorSet, Literal["dim"]] = "keep",
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["grow"], float] = 1.3,
        height: Union[Literal["grow"], float] = 1.3,
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.icon = icon
        self.fill = fill

    def _custom_serialize(self) -> JsonDoc:
        # Get the icon's SVG
        registry = Icon._get_registry()
        svg_source = registry.get_icon_svg(self.icon)

        # Serialize the fill. This isn't automatically handled because it's a
        # Union.
        if isinstance(self.fill, fills.Fill):
            fill = self.fill._serialize(self.session)
        elif isinstance(self.fill, color.Color):
            fill = self.fill.rgba
        else:
            assert isinstance(self.fill, str), f"Unsupported fill type: {self.fill}"
            fill = self.fill

        # Serialize
        return {
            "svgSource": svg_source,
            "fill": fill,
        }


Icon._unique_id = "Icon-builtin"



from typing import *  # type: ignore

from .component_base import FundamentalComponent

__all__ = ["Html"]


class Html(FundamentalComponent):
    """
    Renders HTML.

    `Html` allows you to embed HTML in your app. It takes HTML code as input and
    renders it. If you want to embed an entire website, use the `Website`
    component instead.

    Attributes:
        html: The HTML to render.
    """

    html: str


Html._unique_id = "Html-builtin"



from dataclasses import field
from typing import List, Literal, Mapping, Optional, Union

import rio

from .component_base import Component

__all__ = ["LabeledColumn"]


class LabeledColumn(Component):
    _child_list: List[Component] = field(init=False)

    def __init__(
        self,
        children: Mapping[str, Component],
        *,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["natural", "grow"], float] = "natural",
        height: Union[Literal["natural", "grow"], float] = "natural",
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = children

    @property
    def children(self) -> Mapping[str, Component]:
        return self._children

    @children.setter
    def children(self, children: Mapping[str, Component]) -> None:
        self._children = children
        self._child_list = list(children.values())

    def build(self) -> Component:
        rows = []

        for label, child in self.children.items():
            rows.append(
                [
                    rio.Text(label, align_x=1),
                    child,
                ]
            )

        return rio.Grid(*rows, row_spacing=0.1, column_spacing=0.2)



from __future__ import annotations

from dataclasses import KW_ONLY
from typing import *  # type: ignore

import rio

from . import component_base, progress_circle

__all__ = [
    "Button",
    "IconButton",
]


INITIALLY_DISABLED_FOR = 0.25


class Button(component_base.Component):
    """
    A clickable button.

    The `Button` component allows the user to trigger an action by clicking on
    it. You can use it to trigger a function call, navigate to a different page,
    or perform other actions.

    A similar, but more customizable component is `CustomButton`. Use that if
    you need more control over the button's visuals, e.g. for creating a
    navigation button.

    Attributes:
        child: The text or child component to display inside of the button.

        icon: The name of an icon to display on the button, in the form
            "set/name:variant". See the `Icon` component for details of how
            icons work in Rio.

        shape: The shape of the button. This can be one of:
            - `pill`: A rectangle where the left and right sides are completely
              round.
            - `rounded`: A rectangle with rounded corners.
            - `rectangle`: A rectangle with sharp corners.
            - `circle`: An ellipse. Make sure to control the button's width and
              height to make it a perfect circle.

        style: Controls the button's appearance. This can be one of:
            - `major`: A highly visible button with bold visuals.
            - `minor`: A less visible button that blends into the background.

        color: The color scheme to use for the button.

        is_sensitive: Whether the button should respond to user input.

        is_loading: Whether the button should display a loading indicator. Use
            this to indicate to the user that an action is currently running.

        on_press: Triggered when the user clicks on the button.
    """

    child: Union[rio.Component, str] = ""
    _: KW_ONLY
    icon: Optional[str] = None
    shape: Literal["pill", "rounded", "rectangle"] = "pill"
    style: Literal["major", "minor", "plain"] = "major"
    color: rio.ColorSet = "keep"
    is_sensitive: bool = True
    is_loading: bool = False
    initially_disabled_for: float = INITIALLY_DISABLED_FOR
    on_press: rio.EventHandler[[]] = None

    def build(self) -> rio.Component:
        # Prepare the child
        if self.is_loading:
            if self.color in ("keep", "secondary"):
                progress_color = "primary"
            else:
                progress_color = "secondary"

            child = progress_circle.ProgressCircle(
                size=1.5,
                align_x=0.5,
                margin=0.3,
                color=progress_color,
            )
        elif isinstance(self.child, component_base.Component):
            child = rio.Column(
                self.child,
                align_x=0.5,
            )
        elif self.icon is None:
            child = rio.Text(
                self.child.strip(),
                # Make sure there's no popping when switching between Text & ProgressCircle
                height=1.6,
                margin_x=0.7,
                margin_y=0.3,
            )
        else:
            child = rio.Row(
                rio.Icon(
                    self.icon,
                    height=1.2,
                    width=1.2,
                ),
                rio.Text(
                    self.child.strip(),
                    height=1.5,
                    width="grow",
                ),
                spacing=0.6,
                align_x=0.5,
                # Make sure there's no popping when switching between Text & ProgressCircle
                height=1.6,
                margin_x=0.7,
                margin_y=0.3,
            )

        # Delegate to a HTML Component
        return _ButtonInternal(
            on_press=self.on_press,
            child=child,
            shape=self.shape,
            style=self.style,
            color=self.color,
            is_sensitive=self.is_sensitive,
            is_loading=self.is_loading,
            width=8,
            initially_disabled_for=self.initially_disabled_for,
        )

    def __str__(self) -> str:
        if isinstance(self.child, str):
            text_or_child = f"text:{self.child!r}"
        else:
            text_or_child = f"child:{self.child._id}"

        return f"<Button id:{self._id} {text_or_child}>"


class IconButton(component_base.Component):
    """
    A round, clickable button with shadow.

    The `FloatingActionButton` component is similar to the `Button` component,
    but has a different visual style. It is round and has a shadow, making it
    appear to hover above the rest of the page. It is typically used to make the
    most important action on a page stand out. For example, an e-mail client
    might use a floating action button to compose a new e-mail. A note taking
    app might use it to create a new note.

    The `FloatingActionButton` itself doesn't perform any special layouting.
    Combine it with `rio.Stack` or `rio.Overlay` to make it float above other
    components.
    """

    icon: str
    _: KW_ONLY
    style: Literal["major", "minor", "plain"]
    color: rio.ColorSet
    is_sensitive: bool
    on_press: rio.EventHandler[[]]
    initially_disabled_for: float = INITIALLY_DISABLED_FOR
    size: float

    def __init__(
        self,
        icon: str,
        *,
        style: Literal["major", "minor", "plain"] = "major",
        color: rio.ColorSet = "keep",
        is_sensitive: bool = True,
        on_press: rio.EventHandler[[]] = None,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        size: float = 3.7,
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width="natural",
            height="natural",
            align_x=align_x,
            align_y=align_y,
        )

        self.icon = icon
        self.size = size
        self.style = style
        self.color = color
        self.is_sensitive = is_sensitive
        self.on_press = on_press
        self.initially_disabled_for = INITIALLY_DISABLED_FOR

    def build(self) -> rio.Component:
        return _ButtonInternal(
            on_press=self.on_press,
            child=rio.Icon(
                self.icon,
                height=self.size * 0.65,
                width=self.size * 0.65,
                align_x=0.5,
                align_y=0.5,
            ),
            shape="circle",
            style=self.style,
            color=self.color,
            is_sensitive=self.is_sensitive,
            is_loading=False,
            width=self.size,
            height=self.size,
            initially_disabled_for=self.initially_disabled_for,
        )


class _ButtonInternal(component_base.FundamentalComponent):
    _: KW_ONLY
    on_press: rio.EventHandler[[]]
    child: rio.Component
    shape: Literal["pill", "rounded", "rectangle", "circle"]
    style: Literal["major", "minor", "plain"]
    color: rio.ColorSet
    is_sensitive: bool
    is_loading: bool
    initially_disabled_for: float

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg
        assert msg["type"] == "press", msg

        msg_type: str = msg["type"]
        assert isinstance(msg_type, str), msg_type

        # Is the button sensitive?
        if not self.is_sensitive or self.is_loading:
            return

        # Trigger the press event
        await self.call_event_handler(self.on_press)

        # Refresh the session
        await self.session._refresh()


_ButtonInternal._unique_id = "Button-builtin"



from __future__ import annotations

from typing import *  # type: ignore

import rio

from . import component_base

__all__ = ["Column"]


class Column(component_base.FundamentalComponent):
    """
    A container that lays out its children vertically.

    `Column`s are one of the most common components in Rio. They take any number
    of children and lay them out vertically, with the first one at the top, the
    second one below that, and so on. All components in `Column`s occupy the
    full width of their parent.

    The `Column`'s horizontal counterpart is the `Row`. A similar component, but
    stacking its children in the Z direction, is the `Stack`.

    ### Undefined Space

    Like most containers in `rio`, `Column`s always attempt to allocate all
    available space to their children. In the context of a `Column` though, this
    could easily lead to unexpected results. If more space is available than
    needed, should all children grow? Only the first? Should they grow by equal
    amounts, or proportionally?

    To avoid this ambiguity, `Column`s have a concept of *undefined space*.
    Simply put, **not using all available space is considered an error and
    should be avoided.** `Column`s indicate this by highlighting the extra space
    with unmistakable animated sprites.

    Getting rid of undefined space is easy: Depending on what look you're going
    for, either add a `Spacer` somewhere into your `Column`, assign one of the
    components a `"grow"` value as its height, or set the `Column`'s vertical
    alignment.

    Attributes:
        children: The components to place in this `Column`.

        spacing: How much empty space to leave between two adjacent children. No
            spacing is added before the first child or after the last child.
    """

    children: List[component_base.Component]
    spacing: float

    def __init__(
        self,
        *children: rio.Component,
        spacing: float = 0.0,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["natural", "grow"], float] = "natural",
        height: Union[Literal["natural", "grow"], float] = "natural",
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        assert isinstance(children, tuple), children
        for child in children:
            assert isinstance(child, component_base.Component), child

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = list(children)
        self.spacing = spacing


Column._unique_id = "Column-builtin"



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "Revealer",
    "RevealerChangeEvent",
]

T = TypeVar("T")


@dataclass
class RevealerChangeEvent:
    is_expanded: bool


class Revealer(component_base.FundamentalComponent):
    header: Optional[str]
    content: component_base.Component
    _: KW_ONLY
    is_expanded: bool = False
    on_change: rio.EventHandler[RevealerChangeEvent] = None

    async def _on_state_update(self, delta_state: JsonDoc) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["is_on"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, bool), new_value
            await self.call_event_handler(
                self.on_change,
                RevealerChangeEvent(new_value),
            )

        # Chain up
        await super()._on_state_update(delta_state)


Revealer._unique_id = "Revealer-builtin"



from __future__ import annotations

import logging
from dataclasses import KW_ONLY, dataclass, field
from typing import *  # type: ignore

import rio

from . import common
from .errors import NavigationFailed


@dataclass(frozen=True)
class Page:
    """
    A routable page in a Rio app.

    Rio apps can consist of many pages. You might have a welcome page, a
    settings page, a login, and so on. `Page` components contain all information
    needed to display those pages, as well as to navigate between them.

    A minimal example:

    ```python
    import rio

    app = rio.App(
        build=lambda: rio.Column(
            rio.Text("Welcome to my page!"), rio.PageView(
                width="grow", height="grow",
            ),
        ), pages=[
            rio.Page(
                "", build=lambda: rio.Text("This is the home page"),
            ), rio.Page(
                "subpage", build=lambda: rio.Text("This is a subpage"),
            ),
        ],
    )

    app.run_in_browser()
    ```

    This will display "This is the home page" when navigating to the root URL,
    but "This is a subpage" when navigating to "/subpage". Note that on both
    pages the text "Welcome to my page!" is displayed above the page content.
    That's because it's not part of the `PageView`.

    # TODO: Link to the routing/multipage how-to page

    Attributes:
        page_url: The URL segment at which this page should be displayed. For
            example, if this is "subpage", then the page will be displayed at
            "https://yourapp.com/subpage". If this is "", then the page will be
            displayed at the root URL.

        build: A callback that is called when this page is displayed. It should
            return a Rio component.

        children: A list of child pages. These pages will be displayed when
            navigating to a sub-URL of this page. For example, if this page's
            `page_url` is "page1", and it has a child page with `page_url`
            "page2", then the child page will be displayed at
            "https://yourapp.com/page1/page2".

        guard: A callback that is called before this page is displayed. It
            can prevent users from accessing pages which they are not allowed to
            see. For example, you may want to redirect users to your login page
            if they are trying to access their profile page without being
            logged in.

            The callback should return `None` if the user is allowed to access
            the page, or a string or `rio.URL` if the user should be redirected
            to a different page.
    """

    page_url: str
    build: Callable[[], rio.Component]
    _: KW_ONLY
    children: List["Page"] = field(default_factory=list)
    guard: Optional[Callable[[rio.Session], Union[None, rio.URL, str]]] = None


class PageRedirect(Exception):
    """
    Internal exception solely used internally by `_check_page_guards`. You
    should never see this exception pop up anywhere outside of that function.
    """

    def __init__(self, redirect: rio.URL):
        self.redirect = redirect


def check_page_guards(
    sess: rio.Session,
    target_url_absolute: rio.URL,
) -> Tuple[List[Page], rio.URL]:
    """
    Check whether navigation to the given target URL is possible.

    This finds the pages that would be activated by this navigation and runs
    their guards. If the guards effect a redirect, it instead attempts to
    navigate to the redirect target, and so on.

    Raises `NavigationFailed` if navigation to the target URL is not possible
    because of an error, such as an exception in a guard.

    If the URL points to a page which doesn't exist that is not considered an
    error. The result will still be valid. That is because navigation is
    possible, it's just that some PageViews will display a 404 page.

    This function does not perform any actual navigation. It simply checks
    whether navigation to the target page is possible.
    """

    assert target_url_absolute.is_absolute(), target_url_absolute

    # Is any guard opposed to this page?
    initial_target_url = target_url_absolute
    visited_redirects = {target_url_absolute}
    past_redirects = [target_url_absolute]

    def check_guard(
        pages: Iterable[rio.Page],
        remaining_segments: Tuple[str, ...],
    ) -> List[Page]:
        # Get the page responsible for this segment
        try:
            page_segment = remaining_segments[0]
        except IndexError:
            page_segment = ""

        for page in pages:
            if page.page_url == page_segment:
                break
        else:
            return []

        # Run the guard
        if page.guard is not None:
            try:
                redirect_page = page.guard(sess)
            except Exception as err:
                raise NavigationFailed("Uncaught exception in page guard") from err

            # If a redirect was requested stop recursing
            if isinstance(redirect_page, str):
                redirect_page = rio.URL(redirect_page)

            if isinstance(redirect_page, rio.URL):
                redirect_page = sess.active_page_url.join(redirect_page)

            if redirect_page is not None and redirect_page != target_url_absolute:
                raise PageRedirect(redirect_page)

        # Recurse into the children
        sub_pages = check_guard(page.children, remaining_segments[1:])
        return [page] + sub_pages

    while True:
        # TODO: What if the URL is not a child of the base URL? i.e. redirecting
        #   to a completely different site
        target_url_relative = common.make_url_relative(
            sess._base_url, target_url_absolute
        )

        # Find all pages which would by activated by this navigation, and
        # check their guards
        try:
            page_stack = check_guard(sess.app.pages, target_url_relative.parts)

        # Redirect
        except PageRedirect as err:
            redirect = err.redirect

        # Done
        else:
            return page_stack, target_url_absolute

        assert redirect.is_absolute(), redirect

        # Detect infinite loops and break them
        if redirect in visited_redirects:
            page_strings = [str(page_url) for page_url in past_redirects + [redirect]]
            page_strings_list = "\n -> ".join(page_strings)

            message = f"Rejecting navigation to `{initial_target_url}` because page guards have created an infinite loop:\n\n    {page_strings_list}"
            logging.warning(message)
            raise NavigationFailed(message)

        # Remember that this page has been visited before
        visited_redirects.add(redirect)
        past_redirects.append(redirect)

        # Rinse and repeat
        target_url_absolute = redirect



from __future__ import annotations

import enum
from datetime import timedelta
from typing import *  # type: ignore

import rio

__all__ = [
    "on_create",
    "on_page_change",
    "on_mount",
    "on_unmount",
    "periodic",
]

C = TypeVar("C", bound="rio.Component")
R = TypeVar("R")
SyncOrAsync = Union[R, Awaitable[R]]
SyncOrAsyncNone = TypeVar("SyncOrAsyncNone", bound=SyncOrAsync[None])


class EventTag(enum.Enum):
    ON_CREATE = enum.auto()
    ON_PAGE_CHANGE = enum.auto()
    ON_MOUNT = enum.auto()
    ON_UNMOUNT = enum.auto()
    PERIODIC = enum.auto()


def on_create(handler: Callable[[C], None]) -> Callable[[C], None]:
    """
    Triggered after the component has been created and has access to all of its
    attributes and session. Use this if you need to do any setup that requires
    accessing the component's attributes.

    Accessing attributes in `__init__` doesn't work reliably, because the state
    bindings are only initialized after `__init__` has run.
    """
    handler._rio_event_tag_ = EventTag.ON_CREATE
    return handler


def on_page_change(handler: Callable[[C], R]) -> Callable[[C], R]:
    """
    Triggered whenever the session changes pages.
    """
    handler._rio_event_tag_ = EventTag.ON_PAGE_CHANGE
    return handler


def on_mount(handler: Callable[[C], R]) -> Callable[[C], R]:
    """
    Triggered when the component is added to the widget tree.

    This may be triggered multiple times if the component is removed and then
    re-added.
    """
    handler._rio_event_tag_ = EventTag.ON_MOUNT
    return handler


def on_unmount(handler: Callable[[C], R]) -> Callable[[C], R]:
    """
    Triggered when the component is removed from the widget tree.

    This may be triggered multiple times if the component is removed and then
    re-added.
    """
    handler._rio_event_tag_ = EventTag.ON_UNMOUNT
    return handler


def periodic(
    interval: Union[float, timedelta]
) -> Callable[[Callable[[C], SyncOrAsyncNone]], Callable[[C], SyncOrAsyncNone]]:
    """
    TODO / unfinished / do not use

    This event is triggered repeatedly at a fixed time interval for as long as
    the component exists. The component does not have to be mounted for this
    event to trigger.

    The interval only starts counting after the previous handler has finished
    executing, so the handler will never run twice simultaneously, even if it
    takes longer than the interval to execute.

    Args:
        period: The number of seconds, or timedelta, between each trigger.
    """
    # Convert timedelta to float
    if isinstance(interval, timedelta):
        interval = interval.total_seconds()

    def decorator(
        handler: Callable[[C], SyncOrAsyncNone]
    ) -> Callable[[C], SyncOrAsyncNone]:
        handler._rio_event_tag_ = EventTag.PERIODIC
        handler._rio_event_periodic_interval_ = interval
        return handler

    return decorator



# Re-export dataclass stuff for easy use.
from dataclasses import KW_ONLY, field

# URLs are used as an important datatype within rio. Re-export them for easy
# use.
from yarl import URL

from . import event, font
from .app import *
from .box_style import *
from .color import *
from .common import (
    EventHandler,
    FileInfo,
    ImageLike,
    escape_markdown,
    escape_markdown_code,
)
from .components import *
from .cursor_style import CursorStyle
from .errors import *
from .fills import *
from .routing import Page
from .session import *
from .text_style import *
from .theme import *
from .user_settings_module import *



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

import rio

from . import component_base

__all__ = [
    "NumberInput",
    "NumberInputChangeEvent",
    "NumberInputConfirmEvent",
]


_multiplier_suffixes = {
    "k": 1e3,
    "m": 1e6,
}


@dataclass
class NumberInputChangeEvent:
    value: float


@dataclass
class NumberInputConfirmEvent:
    value: float


class NumberInput(component_base.Component):
    """
    Like `TextInput`, but specifically for inputting numbers.

    `NumberInput` allows the user to enter a number. This is similar to the
    `TextInput` component, but with some goodies for handling numbers. The value
    is automatically parsed and formatted according to the user's locale, and
    you can specify minimum and maximum values to limit the user's input.

    Number inputs go way beyond just parsing the input using `int` or `float`.
    They try their best to understand what the user is trying to enter, and also
    support suffixes such as "k" and "m" to represent thousands and millions
    respectively.

    Attributes:
        value: The number currently entered by the user.

        label: A short text to display next to the text input.

        prefix_text: A short text to display before the text input. Useful for
            displaying currency symbols or other prefixed units.

        suffix_text: A short text to display after the text input. Useful for
            displaying currency names or units.

        minimum: The minimum value the number can be set to.

        maximum: The maximum value the number can be set to.

        decimals: The number of decimals to accept. If the user enters more
            decimals, they will be rounded off. If this value is equal to `0`,
            the input's `value` is guaranteed to be an integer, rather than
            float.

        on_change: Triggered when the user changes the number.

        on_confirm: Triggered when the user explicitly confirms their input,
    """

    value: float = 0
    _: KW_ONLY
    label: str = ""
    prefix_text: str = ""
    suffix_text: str = ""
    minimum: Optional[float] = None
    maximum: Optional[float] = None
    decimals: int = 2
    on_change: rio.EventHandler[NumberInputChangeEvent] = None
    on_confirm: rio.EventHandler[NumberInputConfirmEvent] = None

    def __post_init__(self):
        self._text_input = None

    def _try_set_value(self, raw_value: str) -> bool:
        """
        Parse the given string and update the component's value accordingly.
        Returns `True` if the value was successfully updated, `False` otherwise.
        """

        # Strip the number down as much as possible
        locale = self.session.preferred_locales[0]
        raw_value = raw_value.strip()
        raw_value = raw_value.replace(locale.number_symbols["group"], "")

        # Try to parse the value
        if raw_value:
            # Check for a multiplier suffix
            suffix = raw_value[-1].lower()
            multiplier = 1

            if suffix.isalpha():
                try:
                    multiplier = _multiplier_suffixes[suffix]
                except KeyError:
                    pass
                else:
                    raw_value = raw_value[:-1]

            # Try to parse the number
            try:
                value = float(raw_value.replace(locale.number_symbols["decimal"], "."))
            except ValueError:
                self.value = self.value  # Force the old value to stay
                return False

            # Apply the multiplier
            value *= multiplier

        # If no value was provided choose a reasonable default
        else:
            value = 0

        # Limit the number of decimals
        value = round(value, self.decimals)

        # Clamp the value
        if self.minimum is not None:
            value = max(value, self.minimum)

        if self.maximum is not None:
            value = min(value, self.maximum)

        # Update the value
        self.value = value
        return True

    async def _on_change(self, ev: rio.TextInputChangeEvent) -> None:
        was_updated = self._try_set_value(ev.text)

        if was_updated:
            await self.call_event_handler(
                self.on_change,
                NumberInputChangeEvent(self.value),
            )

    async def _on_confirm(self, ev: rio.TextInputConfirmEvent) -> None:
        was_updated = self._try_set_value(ev.text)

        if was_updated:
            await self.call_event_handler(
                self.on_confirm,
                NumberInputConfirmEvent(self.value),
            )

    def build(self) -> rio.Component:
        # Format the number
        locale = self.session.preferred_locales[0]
        value_str = f"{self.value:.{self.decimals}f}"
        if self.decimals == 0:
            int_str, frac_str = value_str, ""
        else:
            int_str, frac_str = value_str.split(".")

        # Add thousands separators
        groups = []
        while len(int_str) > 3:
            groups.append(int_str[-3:])
            int_str = int_str[:-3]

        groups.append(int_str)
        int_str = locale.number_symbols["group"].join(reversed(groups))

        # Join the strings
        if self.decimals == 0:
            value_str = int_str
        else:
            value_str = int_str + locale.number_symbols["decimal"] + frac_str

        # Build the component
        self._text_input = rio.TextInput(
            text=value_str,
            label=self.label,
            prefix_text=self.prefix_text,
            suffix_text=self.suffix_text,
            on_change=self._on_change,
            on_confirm=self._on_confirm,
        )
        return self._text_input

    async def grab_keyboard_focus(self) -> None:
        if self._text_input is not None:
            await self._text_input.grab_keyboard_focus()



from __future__ import annotations

from typing import *  # type: ignore

import rio

from . import component_base

__all__ = [
    "ClassContainer",
]


class ClassContainer(component_base.FundamentalComponent):
    """
    Component which holds a single child, and applies a list of CSS classes to it.
    This is enough to implement several components, preventing the need to create a
    whole bunch of almost identical JavaScript components.

    This component is only intended for internal use and is not part of the public
    API.
    """

    child: Optional[rio.Component]
    classes: List[str]


ClassContainer._unique_id = "ClassContainer-builtin"



from __future__ import annotations

from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "StyleContext",
]


class StyleContext(component_base.FundamentalComponent):
    child: Optional[rio.Component]
    context: Literal[
        "background",
        "neutral",
        "primary",
        "secondary",
        "success",
        "warning",
        "danger",
    ]

    def _custom_serialize(self) -> JsonDoc:
        return {
            "classes": [
                f"rio-switcheroo-{self.context}",
            ],
        }


# Pretend this is a class container, since applying a switcheroo is all that's
# necessary
StyleContext._unique_id = "ClassContainer-builtin"



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "Dropdown",
    "DropdownChangeEvent",
]

T = TypeVar("T")


@dataclass
class DropdownChangeEvent(Generic[T]):
    value: T


class Dropdown(component_base.FundamentalComponent, Generic[T]):
    """
    A dropdown menu allowing the user to select one of several options.

    Dropdowns present the user with a list of options, allowing them to select
    exactly one. In their default state dropdowns are compact and display the
    currently selected option. When activated, a popup menu appears with a list
    of all available options.

    Attributes:
        options: A mapping from option names to values. The names are displayed
            in the dropdown menu, and the corresponding value is returned when
            the user selects the option. The values must be comparable.

        label: A short text to display next to the dropdown.

        selected_value: The value of the currently selected option.

        on_change: Triggered whenever the user selects an option.

        is_sensitive: Whether the dropdown should respond to user input.
    """

    options: Mapping[str, T]
    _: KW_ONLY
    label: str
    selected_value: T
    on_change: rio.EventHandler[DropdownChangeEvent[T]]
    is_sensitive: bool = True

    def __init__(
        self,
        options: Mapping[str, T],
        *,
        label: str = "",
        selected_value: Optional[T] = None,
        on_change: rio.EventHandler[DropdownChangeEvent[T]] = None,
        is_sensitive: bool = True,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["natural", "grow"], float] = "natural",
        height: Union[Literal["natural", "grow"], float] = "natural",
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        if not options:
            raise ValueError("`Dropdown` must have at least one option.")

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.options = options
        self.label = label
        self.on_change = on_change
        self.is_sensitive = is_sensitive

        # This is an unsafe assignment, because the value could be `None`. This
        # will be fixed in `on_create`, once the state bindings have been
        # initialized.
        self.selected_value = selected_value  # type: ignore

    @rio.event.on_create
    def _on_create(self) -> None:
        # Make sure a value is selected
        if self.selected_value is None:
            self.selected_value = next(iter(self.options.values()))

    def _fetch_selected_name(self) -> str:
        # The frontend works with names, not values. Get the corresponding
        # name.

        # Avoid hammering a potential state binding
        selected_value = self.selected_value

        # Fetch the name
        for name, value in self.options.items():
            if value == selected_value:
                return name
        else:
            raise ValueError(
                f"There is no option with value `{self.selected_value!r}`."
            )

    def _custom_serialize(self) -> JsonDoc:
        result: JsonDoc = {
            "optionNames": list(self.options.keys()),
            "selectedName": self._fetch_selected_name(),
        }

        return result

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        # The frontend works with names, not values. Get the corresponding
        # value.
        try:
            self.selected_value = self.options[msg["name"]]
        except KeyError:
            # Invalid names may be sent due to lag between the frontend and
            # backend. Ignore them.
            return

        # Trigger the event
        await self.call_event_handler(
            self.on_change, DropdownChangeEvent(self.selected_value)
        )

        # Refresh the session
        await self.session._refresh()


Dropdown._unique_id = "Dropdown-builtin"



from __future__ import annotations

import asyncio
import contextlib
import copy
import functools
import inspect
import io
import json
import logging
import secrets
import time
import traceback
import weakref
from datetime import timedelta
from typing import *  # type: ignore

import babel
import fastapi
import pytz
import timer_dict
import uniserde.case_convert
from PIL import Image
from uniserde import Jsonable

import rio

from . import (
    app,
    assets,
    common,
    components,
    debug,
    global_state,
    inspection,
    routing,
    session,
    user_settings_module,
)
from .common import URL
from .errors import AssetError

try:
    import plotly  # type: ignore[missing-import]
except ImportError:
    plotly = None

__all__ = ["AppServer"]


@functools.lru_cache(maxsize=None)
def read_frontend_template(template_name: str) -> str:
    """
    Read a text file from the frontend directory and return its content. The
    results are cached to avoid repeated disk access.
    """
    return (common.GENERATED_DIR / template_name).read_text(encoding="utf-8")


class InitialClientMessage(uniserde.Serde):
    website_url: str
    preferred_languages: List[str]
    timezone: str
    user_settings: Dict[str, Any]
    prefers_light_theme: bool

    window_width: float
    window_height: float


async def _periodically_clean_up_expired_sessions(
    app_server_ref: weakref.ReferenceType[AppServer],
):
    LOOP_INTERVAL = 60 * 15
    SESSION_LIFETIME = 60 * 60

    while True:
        await asyncio.sleep(LOOP_INTERVAL)

        app_server = app_server_ref()
        if app_server is None:
            return

        now = time.monotonic()
        cutoff = now - SESSION_LIFETIME

        for sess in app_server._active_session_tokens.values():
            if sess._last_interaction_timestamp < cutoff:
                sess.close()


class AppServer(fastapi.FastAPI):
    def __init__(
        self,
        app_: app.App,
        running_in_window: bool,
        on_session_start: rio.EventHandler[rio.Session],
        on_session_end: rio.EventHandler[rio.Session],
        default_attachments: Tuple[Any, ...],
        validator_factory: Optional[Callable[[rio.Session], debug.Validator]],
        internal_on_app_start: Optional[Callable[[], None]],
    ):
        super().__init__(lifespan=__class__._lifespan)

        self.app = app_
        self.running_in_window = running_in_window
        self.on_session_start = on_session_start
        self.on_session_end = on_session_end
        self.default_attachments = default_attachments
        self.validator_factory = validator_factory
        self.internal_on_app_start = internal_on_app_start

        # Initialized lazily, when the favicon is first requested.
        self._icon_as_ico_blob: Optional[bytes] = None

        # The session tokens for all active sessions. These allow clients to
        # identify themselves, for example to reconnect in case of a lost
        # connection.
        self._active_session_tokens: Dict[str, rio.Session] = {}

        # All assets that have been registered with this session. They are held
        # weakly, meaning the session will host assets for as long as their
        # corresponding Python objects are alive.
        #
        # Assets registered here are hosted under `/asset/temp-{asset_id}`. In
        # addition the server also permanently hosts other "well known" assets
        # (such as javascript dependencies) which are available under public
        # URLS at `/asset/{some-name}`.
        self._assets: weakref.WeakValueDictionary[
            str, assets.Asset
        ] = weakref.WeakValueDictionary()

        # All pending file uploads. These are stored in memory for a limited
        # time. When a file is uploaded the corresponding future is set.
        self._pending_file_uploads: timer_dict.TimerDict[
            str, asyncio.Future[List[common.FileInfo]]
        ] = timer_dict.TimerDict(default_duration=timedelta(minutes=15))

        # FastAPI

        # self.add_api_route("/app.js.map", self._serve_js_map, methods=["GET"])
        # self.add_api_route("/style.css.map", self._serve_css_map, methods=["GET"])
        self.add_api_route("/rio/favicon.ico", self._serve_favicon, methods=["GET"])
        self.add_api_route(
            "/rio/asset/{asset_id:path}", self._serve_asset, methods=["GET"]
        )
        self.add_api_route(
            "/rio/icon/{icon_name:path}", self._serve_icon, methods=["GET"]
        )
        self.add_api_route(
            "/rio/upload/{upload_token}", self._serve_file_upload, methods=["PUT"]
        )
        self.add_api_websocket_route("/rio/ws", self._serve_websocket)

        # Because this is a single page application, all other routes should
        # serve the index page. The session will determine which components should
        # be shown.
        self.add_api_route(
            "/{initial_route_str:path}", self._serve_index, methods=["GET"]
        )

    @contextlib.asynccontextmanager
    async def _lifespan(self):
        # If running as a server, periodically clean up expired sessions
        if not self.running_in_window:
            assert type(self) is AppServer  # Shut up pyright

            asyncio.create_task(
                _periodically_clean_up_expired_sessions(weakref.ref(self)),
                name="Periodic session cleanup",
            )

        # Trigger the app's startup event
        #
        # This will be done blockingly, so the user can prepare any state before
        # connections are accepted. This is also why it's called before the
        # internal event.
        if self.app._on_app_start is not None:
            try:
                result = self.app._on_app_start(self.app)

                if inspect.isawaitable(result):
                    await result

            # Display and discard exceptions
            except Exception:
                print("Exception in `on_app_start` event handler:")
                traceback.print_exc()

        # Trigger any internal startup event
        if self.internal_on_app_start is not None:
            self.internal_on_app_start()

        yield

    def weakly_host_asset(self, asset: assets.HostedAsset) -> None:
        """
        Register an asset with this server. The asset will be held weakly,
        meaning the server will host assets for as long as their corresponding
        Python objects are alive.

        If another asset with the same id is already hosted, it will be
        replaced.
        """
        self._assets[asset.secret_id] = asset

    def host_asset_with_timeout(self, asset: assets.HostedAsset, timeout: float) -> URL:
        """
        Hosts an asset for a limited time. Returns the asset's url.
        """
        self.weakly_host_asset(asset)

        async def keep_alive():
            await asyncio.sleep(timeout)
            _ = asset

        asyncio.create_task(
            keep_alive(), name=f"Keep asset {asset} alive for {timeout} sec"
        )

        return asset.url

    async def _serve_index(
        self,
        request: fastapi.Request,
        initial_route_str: str,
    ) -> fastapi.responses.HTMLResponse:
        """
        Handler for serving the index HTML page via fastapi.
        """

        # Because Rio apps are single-page, this route serves as the fallback.
        # In addition to legitimate requests for HTML pages, it will also catch
        # a bunch of invalid requests to other resources. To highlight this,
        # throw a 404 if HTML is not explicitly requested.
        #
        # Currently inactive, because this caused issues behind dumb proxies
        # that don't pass on the `accept` header.

        # if not request.headers.get("accept", "").startswith("text/html"):
        #     raise fastapi.HTTPException(
        #         status_code=fastapi.status.HTTP_404_NOT_FOUND,
        #     )

        # Prepare some URL constants
        base_url = rio.URL(str(request.base_url))
        assert base_url.is_absolute(), base_url

        initial_page_url_absolute = rio.URL(str(request.url))
        assert initial_page_url_absolute.is_absolute(), initial_page_url_absolute

        # Create a session instance to hold all of this state in an organized
        # fashion.
        #
        # The session is still missing a lot of values at this point, such as
        # `send_message` and `receive_message`. It will be finished once the
        # websocket connection is established.
        session_token = secrets.token_urlsafe()

        sess = session.Session(
            app_server_=self,
            session_token=session_token,
            base_url=rio.URL(str(request.base_url)),
            initial_page=initial_page_url_absolute,
        )

        self._active_session_tokens[session_token] = sess

        # Run any page guards for the initial page
        try:
            (
                initial_page_instances,
                initial_page_url_absolute,
            ) = routing.check_page_guards(
                sess,
                initial_page_url_absolute,
            )
        except routing.NavigationFailed:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f'Navigation to initial page "{initial_page_url_absolute}" has failed.',
            ) from None

        # Update the session's active page
        sess._active_page_url = initial_page_url_absolute
        sess._active_page_instances = tuple(initial_page_instances)

        # Add any attachments, except for user settings. These are deserialized
        # later on, once the client has sent the initial message.
        for attachment in self.default_attachments:
            if isinstance(attachment, user_settings_module.UserSettings):
                continue

            sess.attachments.add(copy.deepcopy(attachment))

        # Load the templates
        html = read_frontend_template("index.html")

        # Fill in any placeholders
        html = html.replace(
            "{session_token}",
            session_token,
        )

        html = html.replace(
            '"{child_attribute_names}"',
            json.dumps(
                inspection.get_child_component_containing_attribute_names_for_builtin_components()
            ),
        )

        html = html.replace(
            '"{ping_pong_interval}"',
            str(self.app._ping_pong_interval.total_seconds()),
        )

        html = html.replace("{title}", self.app.name)

        # Respond
        return fastapi.responses.HTMLResponse(html)

    async def _serve_js_map(self) -> fastapi.responses.Response:
        """
        Handler for serving the `app.js.map` file via fastapi.
        """
        return fastapi.responses.Response(
            content=read_frontend_template("app.js.map"),
            media_type="application/json",
        )

    async def _serve_css_map(self) -> fastapi.responses.Response:
        """
        Handler for serving the `style.css.map` file via fastapi.
        """
        return fastapi.responses.Response(
            content=read_frontend_template("style.css.map"),
            media_type="application/json",
        )

    async def _serve_favicon(self) -> fastapi.responses.Response:
        """
        Handler for serving the favicon via fastapi, if one is set.
        """
        # If an icon is set, make sure a cached version exists
        if self._icon_as_ico_blob is None and self.app._icon is not None:
            try:
                icon_blob, _ = await self.app._icon.try_fetch_as_blob()

                input_buffer = io.BytesIO(icon_blob)
                output_buffer = io.BytesIO()

                with Image.open(input_buffer) as image:
                    image.save(output_buffer, format="ico")

            except Exception as err:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail="Could not fetch the app's icon.",
                ) from err

            self._icon_as_ico_blob = output_buffer.getvalue()

        # No icon set or fetching failed
        if self._icon_as_ico_blob is None:
            return fastapi.responses.Response(status_code=404)

        # There is an icon, respond
        return fastapi.responses.Response(
            content=self._icon_as_ico_blob,
            media_type="image/x-icon",
        )

    async def _serve_asset(self, asset_id: str) -> fastapi.responses.Response:
        """
        Handler for serving assets via fastapi.

        Some common assets are hosted under permanent, well known URLs under the
        `/asset/{some-name}` path.

        In addition, `HostedAsset` instances are hosted under their secret id
        under the `/asset/temp-{asset_id}` path. These assets are held weakly
        by the session, meaning they will be served for as long as the
        corresponding Python object is alive.
        """

        # Special case: plotly
        #
        # The python plotly library already includes a minified version of
        # plotly.js. Rather than shipping another one, just serve the one
        # included in the library.
        if asset_id == "plotly.min.js" and plotly is not None:
            return fastapi.responses.Response(
                content=plotly.offline.get_plotlyjs(),
                media_type="text/javascript",
            )

        # Well known asset?
        if not asset_id.startswith("temp-"):
            # Construct the path to the target file
            asset_file_path = common.HOSTED_ASSETS_DIR / asset_id

            # Make sure the path is inside the hosted assets directory
            #
            # TODO: Is this safe? Would this allow the client to break out from
            # the directory using tricks such as "../", links, etc?
            asset_file_path = asset_file_path.resolve()
            if common.HOSTED_ASSETS_DIR not in asset_file_path.parents:
                logging.warning(
                    f'Client requested asset "{asset_id}" which is not located inside the hosted assets directory. Somebody might be trying to break out of the assets directory!'
                )
                return fastapi.responses.Response(status_code=404)

            # TODO: Can this check be avoided? `FileResponse` may already be
            # doing it internally anyway.
            if asset_file_path.exists():
                # TODO: Support byte serving
                # https://github.com/tiangolo/fastapi/issues/1240#issuecomment-612485608
                return fastapi.responses.FileResponse(
                    common.HOSTED_ASSETS_DIR / asset_id,
                )

            # No such file
            return fastapi.responses.Response(status_code=404)

        # Get the asset's Python instance. The asset's id acts as a secret, so
        # no further authentication is required.
        try:
            asset = self._assets[asset_id.removeprefix("temp-")]
        except KeyError:
            return fastapi.responses.Response(status_code=404)

        # Fetch the asset's content and respond
        if isinstance(asset, assets.BytesAsset):
            return fastapi.responses.Response(
                content=asset.data,
                media_type=asset.media_type,
            )
        elif isinstance(asset, assets.PathAsset):
            return fastapi.responses.FileResponse(
                asset.path,
                media_type=asset.media_type,
            )
        else:
            assert False, f"Unable to serve asset of unknown type: {asset}"

    async def _serve_icon(self, icon_name: str) -> fastapi.responses.Response:
        """
        Allows the client to request an icon by name. This is not actually the
        mechanism used by the `Icon` component, but allows JavaScript to request
        icons.
        """
        # Get the icon's SVG
        registry = components.Icon._get_registry()

        try:
            svg_source = registry.get_icon_svg(icon_name)
        except AssetError:
            return fastapi.responses.Response(status_code=404)

        # Respond
        return fastapi.responses.Response(
            content=svg_source,
            media_type="image/svg+xml",
        )

    async def _serve_file_upload(
        self,
        upload_token: str,
        file_names: List[str],
        file_types: List[str],
        file_sizes: List[str],
        # If no files are uploaded `files` isn't present in the form data at
        # all. Using a default value ensures that those requests don't fail
        # because of "missing parameters".
        #
        # Lists are mutable, make sure not to modify this value!
        file_streams: List[fastapi.UploadFile] = [],
    ):
        # Try to find the future for this token
        try:
            future = self._pending_file_uploads.pop(upload_token)
        except KeyError:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_400_BAD_REQUEST,
                detail="Invalid upload token.",
            )

        # Make sure the same number of values was received for each parameter
        n_names = len(file_names)
        n_types = len(file_types)
        n_sizes = len(file_sizes)
        n_streams = len(file_streams)

        if n_names != n_types or n_names != n_sizes or n_names != n_streams:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Inconsistent number of files between the different message parts.",
            )

        # Parse the file sizes
        parsed_file_sizes: List[int] = []
        for file_size in file_sizes:
            try:
                parsed = int(file_size)
            except ValueError:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Invalid file size.",
                )

            if parsed < 0:
                raise fastapi.HTTPException(
                    status_code=fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Invalid file size.",
                )

            parsed_file_sizes.append(parsed)

        # Complete the future
        future.set_result(
            [
                common.FileInfo(
                    name=file_names[ii],
                    size_in_bytes=parsed_file_sizes[ii],
                    media_type=file_types[ii],
                    _contents=await file_streams[ii].read(),
                )
                for ii in range(n_names)
            ]
        )

        return fastapi.responses.Response(status_code=fastapi.status.HTTP_200_OK)

    async def _serve_websocket(
        self,
        websocket: fastapi.WebSocket,
        sessionToken: str,
    ):
        """
        Handler for establishing the websocket connection and handling any
        messages.
        """
        # Blah, naming conventions
        session_token = sessionToken
        del sessionToken

        # Accept the socket
        await websocket.accept()

        # Look up the session token. If it is valid the session's duration is
        # refreshed so it doesn't expire. If the token is not valid, don't
        # accept the websocket.
        try:
            sess = self._active_session_tokens[session_token]
        except KeyError:
            # Inform the client that it's session token is unknown and request a
            # refresh
            await websocket.send_json(
                {
                    "jsonrpc": "2.0",
                    "method": "invalidSessionToken",
                    "params": {},
                }
            )

            await websocket.close(
                3000,  # Custom error code
                "Invalid session token.",
            )
            return

        # Optionally create a validator
        validator_instance = (
            None if self.validator_factory is None else self.validator_factory(sess)
        )

        # Create a function for sending messages to the frontend. This function
        # will also pipe the message to the validator if one is present.
        if self.validator_factory is None:

            async def receive_message() -> uniserde.Jsonable:
                # Refresh the session's duration
                self._active_session_tokens[session_token] = sess

                # Fetch a message
                return await websocket.receive_json()

            sess._send_message = websocket.send_json  # type: ignore
            sess._receive_message = receive_message

        else:

            async def send_message(msg: uniserde.Jsonable) -> None:
                assert isinstance(validator_instance, debug.Validator)
                validator_instance.handle_outgoing_message(msg)
                await websocket.send_json(msg)

            async def receive_message() -> uniserde.Jsonable:
                assert isinstance(validator_instance, debug.Validator)
                # Refresh the session's duration
                self._active_session_tokens[session_token] = sess

                # Fetch a message
                msg = await websocket.receive_json()
                validator_instance.handle_incoming_message(msg)
                return msg

            sess._send_message = send_message
            sess._receive_message = receive_message

        # Check if this is a reconnect
        if hasattr(sess, "window_width"):
            init_coro = sess._send_reconnect_message()
        else:
            await self._finish_session_initialization(sess, websocket, session_token)

            # Trigger a refresh. This will also send the initial state to the
            # frontend.
            init_coro = sess._refresh()

        try:
            # This is done in a task because the server is not yet running, so
            # the method would never receive a response, and thus would hang
            # indefinitely.
            sess.create_task(init_coro, name=f"Session {sess} init")

            # Serve the socket
            await sess.serve()

        # Don't spam the terminal just because a client disconnected
        except fastapi.WebSocketDisconnect:
            pass

        finally:
            sess.close()

    async def _finish_session_initialization(
        self,
        sess: session.Session,
        websocket: fastapi.WebSocket,
        session_token: str,
    ) -> None:
        # Upon connecting, the client sends an initial message containing
        # information about it. Wait for that, but with a timeout - otherwise
        # evildoers could overload the server with connections that never send
        # anything.
        initial_message_json: Jsonable = await asyncio.wait_for(
            websocket.receive_json(),
            timeout=60,
        )
        initial_message = InitialClientMessage.from_json(initial_message_json)  # type: ignore

        sess.window_width = initial_message.window_width
        sess.window_height = initial_message.window_height

        # Parse the preferred locales
        preferred_locales: List[babel.Locale] = []

        for raw_locale_string in initial_message.preferred_languages:
            try:
                preferred_locales.append(
                    babel.Locale.parse(
                        raw_locale_string.replace("-", "_"),
                    )
                )
            except ValueError:
                pass

        if not preferred_locales:
            preferred_locales.append(babel.Locale.parse("en"))

        sess.preferred_locales = tuple(preferred_locales)

        # Parse the timezone
        try:
            sess.timezone = pytz.timezone(initial_message.timezone)
        except pytz.UnknownTimeZoneError:
            logging.warning(
                f'Client sent unknown timezone "{initial_message.timezone}". Using UTC instead.'
            )
            sess.timezone = pytz.UTC

        # Publish the external URL via the session
        sess.external_url = (
            None if self.running_in_window else initial_message.website_url
        )

        # Set the theme according to the user's preferences
        theme = self.app._theme
        if isinstance(theme, tuple):
            if initial_message.prefers_light_theme:
                theme = theme[0]
            else:
                theme = theme[1]

        await sess._apply_theme(theme)

        # Deserialize the user settings
        await sess._load_user_settings(initial_message.user_settings)

        # Create the root component. The root component is a non-fundamental component,
        # because that has many advantages:
        # 1. Every component except for the root component itself has a valid builder
        # 2. The JS code is simpler because the root component can't have an
        #    alignment or margin
        # 3. Children of non-fundamental components are automatically initialized
        #    correctly, so we don't need to duplicate that logic here
        global_state.currently_building_component = None
        global_state.currently_building_session = sess

        try:
            sess._root_component = RootContainer(self.app._build)
        finally:
            global_state.currently_building_session = None

        # Trigger the `on_session_start` event.
        #
        # Note: Since this event is often used for initialization, like adding
        # attachments, we actually wait for it to finish before continuing.
        await sess._call_event_handler(self.on_session_start, sess)


class RootContainer(components.Component):
    build_function: Callable[[], components.Component]

    def build(self) -> rio.Component:
        return self.build_function()



from __future__ import annotations

from typing import *  # type: ignore

import rio

from . import component_base

__all__ = ["Stack"]


class Stack(component_base.FundamentalComponent):
    """
    A container that stacks its children in the Z direction.

    `Stacks` are similar to rows and columns, but they stack their children in
    the Z direction instead of the X or Y direction. In other words, the stack's
    children overlap each other, with the first one at the bottom, the second
    one above that, and so on.

    Attributes:
        children: The components to place in this `Stack`.
    """

    children: List[rio.Component]

    def __init__(
        self,
        *children: rio.Component,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["natural", "grow"], float] = "natural",
        height: Union[Literal["natural", "grow"], float] = "natural",
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        for child in children:
            assert isinstance(child, component_base.Component), child

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.children = list(children)


Stack._unique_id = "Stack-builtin"



import logging
from dataclasses import dataclass
from pathlib import Path
from typing import *  # type: ignore

import revel
import toml
import uniserde
from revel import *  # type: ignore
from revel import fatal

__all__ = [
    "RioProject",
]

T = TypeVar("T")


class RioProject:
    def __init__(self, file_path: Path, toml_dict: uniserde.JsonDoc):
        # Path to the `rio.toml` file. May or may not exist
        self._file_path = file_path

        # All of the data from the `rio.toml` file
        self._toml_dict = toml_dict

        # Which keys have been modified and thus must be written back to the
        # `rio.toml` file
        self._dirty_keys: Set[str] = set()

    def _get_key(
        self,
        key_name: str,
        type: Type[T],
        default_value: Any,
    ) -> T:
        # Try to fetch the value from the toml file
        try:
            value = self._toml_dict[key_name]

        # There is no value, use the default
        except KeyError:
            self._toml_dict[key_name] = default_value
            self._dirty_keys.add(key_name)
            return default_value

        # Make sure the value is the correct type
        if not isinstance(value, type):
            fatal(
                f"`rio.toml` contains an invalid value for `{key_name}`: expected {type}, got {type(value)}",
                status_code=1,
            )

        # Done
        return value

    @property
    def main_module(self) -> str:
        return self._get_key("main_module", str, "TODO")

    @property
    def fastapi_app_variable(self) -> str:
        return self._get_key("fastapi_app_variable", str, "fastapi_app")

    @staticmethod
    def try_load() -> "RioProject":
        """
        Best-effort attempt to locate the project directory and load the
        `rio.toml` file.
        """
        # Search upward until a `rio.toml` is found
        root_search_dir = Path.cwd()
        project_dir = root_search_dir

        while True:
            # Is this the project directory?
            rio_toml_path = project_dir / "rio.toml"

            if rio_toml_path.exists():
                break

            # Go up a directory, if possible
            parent_dir = project_dir.parent

            if parent_dir == project_dir:
                project_dir = root_search_dir
                break

            project_dir = parent_dir

        # Try to load the toml file
        rio_toml_path = project_dir / "rio.toml"
        logging.debug(f"Loading `{rio_toml_path}`")

        try:
            rio_toml_dict = toml.load(rio_toml_path)

        # No such file. Offer to create one
        except FileNotFoundError:
            warning(
                f"You don't appear to be inside of a Rio project. Would you like to create one?"
            )

            if not revel.select_yes_no("", default_value=True):
                fatal(
                    f"Couldn't find any a `rio.toml` file.",
                    status_code=1,
                )

            rio_toml_dict = {}

        # Anything OS related
        except OSError as e:
            fatal(
                f"Cannot read `{rio_toml_path}`: {e}",
                status_code=1,
            )

        # Invalid syntax
        except toml.TomlDecodeError as e:
            fatal(
                f"There is a syntax error in the `rio.toml` file: {e}",
                status_code=1,
            )

        # Done, return the project
        return RioProject(
            file_path=rio_toml_path,
            toml_dict=rio_toml_dict,  # type: ignore
        )

    def __enter__(self) -> "RioProject":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        # Make sure to write any changes back to the `rio.toml` file
        if self._dirty_keys:
            self.write()

    def write(self) -> None:
        """
        Write any changes back to the `rio.toml` file.
        """
        logging.debug(f"Writing `{self._file_path}`")

        # Make sure the parent directory exists
        self._file_path.parent.mkdir(parents=True, exist_ok=True)

        # Write the file
        try:
            with self._file_path.open("w") as f:
                toml.dump(self._toml_dict, f)

        except OSError as e:
            fatal(
                f"Couldn't write `{self._file_path}`: {e}",
                status_code=1,
            )



from __future__ import annotations

from dataclasses import KW_ONLY, field
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "CustomListItem",
    "SingleLineListItem",
]


class CustomListItem(component_base.FundamentalComponent):
    child: component_base.Component
    _: KW_ONLY
    on_press: rio.EventHandler[[]] = None

    def _custom_serialize(self) -> JsonDoc:
        return {
            "pressable": self.on_press is not None,
        }

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg
        assert msg["type"] == "press", msg

        msg_type: str = msg["type"]
        assert isinstance(msg_type, str), msg_type

        if self.on_press is None:
            return

        # Trigger the press event
        await self.call_event_handler(self.on_press)

        # Refresh the session
        await self.session._refresh()


class SingleLineListItem(component_base.Component):
    text: str
    _: KW_ONLY
    secondary_text: str = ""
    image_or_icon: Optional[component_base.Component] = None
    actions: List[component_base.Component] = field(default_factory=list)
    on_press: rio.EventHandler[[]] = None

    def build(self) -> component_base.Component:
        children = []

        # Image or Icon
        if self.image_or_icon is not None:
            children.append(self.image_or_icon)

        # Main content (text)
        text_children = [
            rio.Text(
                self.text,
                align_x=0,
            )
        ]

        if self.secondary_text:
            text_children.append(
                rio.Text(
                    self.secondary_text,
                    multiline=True,
                    style="dim",
                    align_x=0,
                )
            )

        if len(text_children) == 1:
            children.append(text_children[0])
        else:
            children.append(
                rio.Column(
                    *text_children,
                    spacing=0.5,
                )
            )

        # Space

        children.append(rio.Spacer())

        # Actions, if any
        children.extend(self.actions)

        # Combine everything
        return CustomListItem(
            child=rio.Row(
                *children,
                spacing=1,
                width="grow",
            ),
            on_press=self.on_press,
        )


CustomListItem._unique_id = "ListItem-builtin"



from __future__ import annotations

from dataclasses import dataclass
from typing import *  # type: ignore

from typing_extensions import Self
from uniserde import Jsonable

import rio

from . import color, self_serializing, session

__all__ = [
    "BoxStyle",
]


@dataclass(frozen=True)
class BoxStyle(self_serializing.SelfSerializing):
    fill: rio.Fill
    stroke_color: rio.Color
    stroke_width: float
    corner_radius: Tuple[float, float, float, float]
    shadow_color: rio.Color
    shadow_radius: float
    shadow_offset_x: float
    shadow_offset_y: float

    def __init__(
        self,
        *,
        fill: rio.FillLike,
        stroke_color: rio.Color = color.Color.BLACK,
        stroke_width: float = 0.0,
        corner_radius: Union[float, Tuple[float, float, float, float]] = 0.0,
        shadow_color: rio.Color = color.Color.BLACK,
        shadow_radius: float = 0.0,
        shadow_offset_x: float = 0.0,
        shadow_offset_y: float = 0.0,
    ):
        fill = rio.Fill._try_from(fill)

        if isinstance(corner_radius, (int, float)):
            corner_radius = (
                corner_radius,
                corner_radius,
                corner_radius,
                corner_radius,
            )

        vars(self).update(
            fill=fill,
            stroke_color=stroke_color,
            stroke_width=stroke_width,
            corner_radius=corner_radius,
            shadow_color=shadow_color,
            shadow_radius=shadow_radius,
            shadow_offset_x=shadow_offset_x,
            shadow_offset_y=shadow_offset_y,
        )

    def replace(
        self,
        *,
        fill: Optional[rio.FillLike] = None,
        stroke_color: Optional[rio.Color] = None,
        stroke_width: Optional[float] = None,
        corner_radius: Optional[Union[float, Tuple[float, float, float, float]]] = None,
        shadow_color: Optional[rio.Color] = None,
        shadow_radius: Optional[float] = None,
        shadow_offset_x: Optional[float] = None,
        shadow_offset_y: Optional[float] = None,
    ) -> Self:
        if fill is not None:
            fill = rio.Fill._try_from(fill)

        if isinstance(corner_radius, (int, float)):
            corner_radius = (
                corner_radius,
                corner_radius,
                corner_radius,
                corner_radius,
            )

        return BoxStyle(
            fill=self.fill if fill is None else fill,
            # Stroke rio.Color
            stroke_color=self.stroke_color if stroke_color is None else stroke_color,
            # Stroke Width
            stroke_width=self.stroke_width if stroke_width is None else stroke_width,
            # Corner Radius
            corner_radius=self.corner_radius
            if corner_radius is None
            else corner_radius,
            # shadow
            shadow_color=self.shadow_color if shadow_color is None else shadow_color,
            shadow_radius=self.shadow_radius
            if shadow_radius is None
            else shadow_radius,
            shadow_offset_x=self.shadow_offset_x
            if shadow_offset_x is None
            else shadow_offset_x,
            shadow_offset_y=self.shadow_offset_y
            if shadow_offset_y is None
            else shadow_offset_y,
        )

    def _serialize(self, sess: session.Session) -> Jsonable:
        return {
            "fill": self.fill._serialize(sess),
            "strokeColor": self.stroke_color.rgba,
            "strokeWidth": self.stroke_width,
            "cornerRadius": self.corner_radius,
            "shadowColor": self.shadow_color.rgba,
            "shadowRadius": self.shadow_radius,
            "shadowOffset": (self.shadow_offset_x, self.shadow_offset_y),
        }



from __future__ import annotations

from dataclasses import KW_ONLY, field
from typing import *  # type: ignore

import rio

from . import component_base

__all__ = [
    "CustomButton",
]


class CustomButton(component_base.Component):
    """
    A clickable button with customizable visuals.

    Most buttons in apps should fit the main Theme and are well served by the
    `Button` component. However, sometimes you need more control over the
    button's appearance. For example, you might want to create a navigation
    button that integrates well with the rest of the navigation bar.

    For those use cases, the `CustomButton` component is available. It gives you
    full control about the button's appearance, while still providing the
    same functionality as regular buttons.

    Attributes:
        text: The text to display on the button.

        is_sensitive: Whether the button should respond to user input.

        style_default: The background style to use when the button is not
            hovered or pressed.

        style_hover: The background style to use when the cursor is hovering
            above the button.

        style_press: The background style to use when the button is pressed.

        style_insensitive: The background style to use when the button is
            not sensitive.

        text_style_default: The text style to use when the button is not
            hovered or pressed.

        text_style_hover: The text style to use when the cursor is hovering
            above the button.

        text_style_press: The text style to use when the button is pressed.

        text_style_insensitive: The text style to use when the button is
            not sensitive.

        transition_speed: How many seconds it takes for the button to transition
            between styles.

        ripple: Whether the button should display a Material Design ripple
            effect when hovered or clicked.

        on_press: Triggered when the user clicks on the button.
    """

    text: str
    _: KW_ONLY
    is_sensitive: bool = True
    style_default: rio.BoxStyle
    style_hover: rio.BoxStyle
    style_press: rio.BoxStyle
    style_insensitive: rio.BoxStyle
    text_style_default: rio.TextStyle
    text_style_hover: rio.TextStyle
    text_style_press: rio.TextStyle
    text_style_insensitive: rio.TextStyle
    transition_speed: float
    ripple: bool = True
    on_press: rio.EventHandler[[]] = None

    # Internals
    _is_pressed: bool = field(init=False, default=False)
    _is_entered: bool = field(init=False, default=False)

    def _on_mouse_enter(self, event: rio.MouseEnterEvent) -> None:
        self._is_entered = True

    def _on_mouse_leave(self, event: rio.MouseLeaveEvent) -> None:
        self._is_entered = False

    def _on_mouse_down(self, event: rio.MouseDownEvent) -> None:
        # Only react to left mouse button
        if event.button != rio.MouseButton.LEFT:
            return

        self._is_pressed = True

    async def _on_mouse_up(self, event: rio.MouseUpEvent) -> None:
        # Only react to left mouse button, and only if sensitive
        if event.button != rio.MouseButton.LEFT or not self.is_sensitive:
            return

        await self.call_event_handler(self.on_press)

        self._is_pressed = False

    def build(self) -> rio.Component:
        # If not sensitive, use the insensitive style
        if not self.is_sensitive:
            style = self.style_insensitive
            text_style = self.text_style_insensitive
            hover_style = None

        # If pressed use the press style
        elif self._is_pressed:
            style = self.style_press
            text_style = self.text_style_press
            hover_style = None

        # Otherwise use the regular styles
        else:
            style = self.style_default
            hover_style = self.style_hover
            text_style = (
                self.text_style_hover if self._is_entered else self.text_style_default
            )

        # Prepare the child
        child = rio.Text(
            self.text,
            style=text_style,
            margin=0.3,
        )

        return rio.MouseEventListener(
            rio.Rectangle(
                child=child,
                style=style,
                hover_style=hover_style,
                transition_time=self.transition_speed,
                cursor=rio.CursorStyle.POINTER
                if self.is_sensitive
                else rio.CursorStyle.DEFAULT,
            ),
            on_mouse_enter=self._on_mouse_enter,
            on_mouse_leave=self._on_mouse_leave,
            on_mouse_down=self._on_mouse_down,
            on_mouse_up=self._on_mouse_up,
        )



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import KW_ONLY, dataclass
from typing import *  # type: ignore

from uniserde import JsonDoc

import rio

from . import component_base

__all__ = [
    "SwitcherBarChangeEvent",
    "SwitcherBar",
]

T = TypeVar("T")


@dataclass
class SwitcherBarChangeEvent(Generic[T]):
    value: T


class SwitcherBar(component_base.FundamentalComponent, Generic[T]):
    options: Mapping[str, T]
    _: KW_ONLY
    color: rio.ColorSet
    orientation: Literal["horizontal", "vertical"]
    spacing: float
    selected_value: T
    on_change: rio.EventHandler[SwitcherBarChangeEvent[T]]

    def __init__(
        self,
        options: Mapping[str, T],
        *,
        color: rio.ColorSet = "keep",
        orientation: Literal["horizontal", "vertical"] = "horizontal",
        spacing: float = 1.0,
        selected_value: Optional[T] = None,
        on_change: rio.EventHandler[SwitcherBarChangeEvent[T]] = None,
        key: Optional[str] = None,
        margin: Optional[float] = None,
        margin_x: Optional[float] = None,
        margin_y: Optional[float] = None,
        margin_left: Optional[float] = None,
        margin_top: Optional[float] = None,
        margin_right: Optional[float] = None,
        margin_bottom: Optional[float] = None,
        width: Union[Literal["natural", "grow"], float] = "natural",
        height: Union[Literal["natural", "grow"], float] = "natural",
        align_x: Optional[float] = None,
        align_y: Optional[float] = None,
    ):
        if not options:
            raise ValueError("`SwitcherBar` must have at least one option.")

        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            width=width,
            height=height,
            align_x=align_x,
            align_y=align_y,
        )

        self.options = options
        self.color = color
        self.orientation = orientation
        self.spacing = spacing
        self.on_change = on_change

        # This is an unsafe assignment, because the value could be `None`. This
        # will be fixed in `on_create`, once the state bindings have been
        # initialized.
        self.selected_value = selected_value  # type: ignore

    @rio.event.on_create
    def _on_create(self) -> None:
        # Make sure a value is selected
        if self.selected_value is None:
            self.selected_value = next(iter(self.options.values()))

    def _fetch_selected_name(self) -> str:
        # The frontend works with names, not values. Get the corresponding
        # name.

        # Avoid hammering a potential state binding
        selected_value = self.selected_value

        # Fetch the name
        for name, value in self.options.items():
            if value == selected_value:
                return name
        else:
            raise ValueError(
                f"There is no option with value `{self.selected_value!r}`."
            )

    def _custom_serialize(self) -> JsonDoc:
        thm = self.session.theme
        result = {
            "optionNames": list(self.options.keys()),
            "selectedName": self._fetch_selected_name(),
            "color": thm._serialize_colorset(self.color),
        }

        return result

    async def _on_message(self, msg: Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        # The frontend works with names, not values. Get the corresponding
        # value.
        try:
            self.selected_value = self.options[msg["name"]]
        except KeyError:
            # Invalid names may be sent due to lag between the frontend and
            # backend. Ignore them.
            return

        # Trigger the event
        await self.call_event_handler(
            self.on_change, SwitcherBarChangeEvent(self.selected_value)
        )

        # Refresh the session
        await self.session._refresh()


SwitcherBar._unique_id = "SwitcherBar-builtin"



from typing import *  # type: ignore

import rio


class HistoryItem(rio.Component):
    from_string: str
    to_string: str
    title: str
    organization: str
    details: str
    icon: str

    is_expanded: bool = False

    def on_press(self) -> None:
        self.is_expanded = not self.is_expanded

    def build(self) -> rio.Component:
        return rio.Card(
            rio.Row(
                rio.Icon(
                    self.icon,
                    width=3,
                    height=3,
                    margin_right=2,
                    align_x=0,
                    align_y=0,
                ),
                rio.Column(
                    rio.Text(
                        f"{self.from_string} - {self.to_string}",
                        align_x=0,
                        style=rio.TextStyle(fill=rio.Color.GREY),
                    ),
                    rio.Text(
                        f"{self.title} at {self.organization}",
                        style="heading3",
                        align_x=0,
                    ),
                    rio.Revealer(
                        None,
                        rio.Text(
                            self.details,
                            multiline=True,
                            align_x=0,
                        ),
                        is_expanded=self.is_expanded,
                    ),
                    spacing=0.3,
                    width="grow",
                    align_y=0,
                ),
            ),
            on_press=self.on_press,
        )


class History(rio.Component):
    def build(self) -> rio.Component:
        return rio.Column(
            rio.Text(
                "Professional Experience",
                style="heading2",
                margin_top=3,
            ),
            HistoryItem(
                "2021",
                "Present",
                "AI Researcher",
                "DataWorks Inc.",
                "Developed and deployed machine learning models to improve the "
                "accuracy of the company's recommender system. Collaborated "
                "with a team of 10 researchers to implement a novel algorithm "
                "for predicting user behavior.",
                "science",
            ),
            HistoryItem(
                "2018",
                "2021",
                "Data Analyst",
                "Insight Technologies",
                "Conducted in-depth data analysis and implemented machine "
                "learning algorithms to improve predictive models, resulting "
                "in a 15% increase in accuracy. Collaborated with "
                "cross-functional teams to develop and deploy scalable AI "
                "solutions for optimizing business operations.",
                "monitoring",
            ),
            rio.Text(
                "Education",
                style="heading2",
                margin_top=3,
            ),
            HistoryItem(
                "2015",
                "2018",
                "Master in Data Science",
                "Carnegie Mellon University, Pittsburgh, PA",
                'Thesis: "Predictive Modeling in Financial Markets using '
                'Machine Learning Techniques"',
                "school",
            ),
            HistoryItem(
                "2012",
                "2015",
                "Bachelor in Software Engineering",
                "University of Texas, Austin, TX",
                'Thesis: "Design and Implementation of an Intelligent Traffic '
                'Control System"',
                "school",
            ),
            spacing=2,
        )
